//
//									CROSSLOG
// Visualization program for geophysical well data stored in LAS format 
// http://www.epgeology.com/petrophysics-f23/las-file-standards-t2608.html
// Display a selected curve, plot its histogram, cross-plot log data
//
// Copyright 2011-2015 by Yuri Ganshin, CMI, University of Wyoming, Laramie.
// http://www.uwyo.edu/cmi/
//
// This program is free sofware released using the Fast Light Tool Kit (FLTK).
// For conditions of distribution and use, see the license agreement in 
//     http://www.fltk.org/COPYING.php
// As a general rule, "free software" means that the user have the freedom
// to run, copy, distribute, study, modify and improve the software.
//
// Please report all bugs and problems on the following page:
// http://www.uwyo.edu/cmi/
//     
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <iostream>
//
#include <FL/Fl.H>
#include <FL/Fl_Window.H>
#include <FL/Fl_Double_Window.H>
#include <FL/fl_draw.H>	
#include <FL/Fl_Menu_Bar.H>
#include <FL/Fl_Native_File_Chooser.H>
#include <FL/Fl_Int_Input.H>
#include <FL/Fl_Value_Slider.H>
#include <FL/Fl_Spinner.H>
#include <FL/Fl_Float_Input.H>
#include <FL/Fl_Input_Choice.H>
#include <FL/Fl_Check_Button.H>
#include <FL/fl_show_colormap.H>
#include <FL/Fl_Output.H>
#include <FL/Fl_Button.H>
#include <FL/Fl_Box.H>
#include <FL/fl_message.H>
#include <FL/fl_ask.H> // for fl_alert() and fl_message()
#include <math.h>

using namespace std;

#define MARGIN		60
#define RIM			12
#define MAX_CURVE	80
#define FLT_MAX	1E+37
#define FLT_MIN	-1E+37
#define HYST_BINS	100	// Fixed number of bins in a hystogram
//
#define PALSZ		11
#define PALINC		0.1	// Increment of values within a palette
//
#ifndef M_PI
#define M_PI 3.14159265359
#endif	// Tells CPP to define M_PI if it isn't defined already
//
////////////////////////////////////////////////////////////////
/****               List of GLOBAL variables               ****/
FILE *in_fp = NULL;
char inpname[200],c_word[4];
const char *unit[] = {"meter","foot","unknown"};
float no_value=-9999;			// Absent values (gaps in log data) 
int chosen=-1, dep_unit=2;		// Depth unit: 0=meter, 1=foot, -1=unknown
int Ncur=-1;					// Number of curves in file
int Ndat=-1;					// Number of samples in a curve
//
int win_w, win_h, hist_w, hist_h, xplot_w, xplot_h;
//
float depth_sta, depth_end, depth_step;  // for the whole log length
int samp_sta, samp_fin, samp_vis;  // for selected log interval
float dep_sta, dep_end;			// for displayed part of the log 
float min_val, max_val;			// Min-Max of displayed curves
float *logs, *mini, *maxi;		// Min-Max for all curves
//
char curve[MAX_CURVE][22],cgbuf[40];
int pt_size=2, line_draw=0;
int curve_no = 1;					// default values
int cur1=1, cur2=1, cur3=1, cur4=1;	// curves to cross-plot
float min1,min2,min3,min4,max1,max2,max3,max4;
Fl_Color bg_color = FL_WHITE;
Fl_Color colr = 0;					// Curve color

///////////////////  Color Palettes  ////////////////////
//		array size = PALSZ*4,  palette size = PALSZ
//
float *pal;				// Palette container
//
float RGB_pal[] =		// Red-Green-Blue 
{	0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0,
	125, 255, 255, 255, 125,   0,   0,   0,   0,   0,   0,
	  0,   0, 125, 255, 255, 175, 255, 255, 125,   0,   0,
	  0,   0,   0,   0,   0,   0, 125, 255, 255, 255, 125};
//
float BGR_pal[] =		// Blue-Green-Red
{	0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0,
	  0,   0,   0,   0,   0,   0, 125, 255, 255, 255, 125,
	  0,   0, 125, 255, 255, 175, 255, 255, 125,   0,   0,
	125, 255, 255, 255, 125,   0,   0,   0,   0,   0,   0};
//
float BYD_pal[] =		// Blue-Yellow-Dark
{	0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0,
	  0,   0,   0,  32,   0, 255, 210, 244, 139,  92,  75,
	  0,   0, 255, 178, 205, 255, 180, 164,  69,  64,   0,
	175, 255, 255, 170,   0,   0, 140,  96,  19,  51,   0};
//
float MGR_pal[] =		// Magenta-Green-Red
{	0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0,
	 75, 139,  85,   0,   0,   0,   0, 255, 255, 255, 142,
	  0,   0,  26,   0, 245, 139, 255, 255, 127,   0,  35,
	 75, 139, 139, 205, 255,  69,   0,   0,   0,   0,  35};
//
float BWR_pal[] =		// Blue-Gray-Red
{	0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0,
	  0,   0,   0,   0, 160, 201, 201, 255, 255, 255, 125,
	  0,   0, 125, 255, 180, 201, 170, 255, 125,   0,   0,
	125, 255, 255, 255, 201, 201, 125,   0,   0,   0,   0};
//
////////////////////////////////////////////////////////////////
/****            List of Function Prototypes               ****/

void read_file_info(const char *name);
float conv3(float *dat,int NS1,int NS2,int index);
void pal_interp(float [], float, short []);

////////////////////////////////////////////////////////////////
/****         Global pointers to the GUI objects           ****/
////////////////////////////////////////////////////////////////

Fl_Double_Window *curve_window, *histo_window, *X_window;
Fl_Window *opts_window;
Fl_Float_Input *min_inp,*max_inp,*dep_sta_inp,*dep_end_inp;
Fl_Float_Input *min1_inp, *min2_inp, *min3_inp, *min4_inp;
Fl_Float_Input *max1_inp, *max2_inp, *max3_inp, *max4_inp;
Fl_Button *apply, *x_plot, *histo, *choose_color, *end_this;
Fl_Check_Button *connected;

/////////////////////////////////////////////////////////////////////
/****          List of Widgets and Drawing functions            ****/
/////////////////////////////////////////////////////////////////////

/**********     File-choosing Menu    ************/
class File_menu : public Fl_Window {
	Fl_Native_File_Chooser *fc;
	
	void read_data(const char *filename)	// Process the chosen file
	{
		strcpy(inpname,filename);
		chosen=1;			// will wait while(chosen<0)
	}
	
	// Handle an 'Open' request from the menu
	static void open_cb(Fl_Widget *w, void *v) {
		File_menu *choice = (File_menu*)v;
		choice->fc->title("Open");
		choice->fc->type(Fl_Native_File_Chooser::BROWSE_FILE);
		switch ( choice->fc->show() ) {
			case -1: break;	// Error
			case  1: break;	// Cancel
			default:		// choice->fc->show() returns 0
				choice->read_data(choice->fc->filename());
				// choice->deactivate();				
				break;
		}
	}
	
	static void quit_cb(Fl_Widget *w, void *v) {
		exit(0);
	}
	
public:
	File_menu() : Fl_Window(500,300,320,160,"Choose the file") {
		Fl_Menu_Bar *menu = new Fl_Menu_Bar(0,0,300,25);
		menu->add("&File/&Open",  FL_COMMAND+'o', open_cb, (void*)this);
		menu->add("&File/&Quit",  FL_COMMAND+'q', quit_cb);
		
		// Describe the chooser
		Fl_Box *box = new Fl_Box(20,45,w()-40,h()-65);
		box->color(FL_GRAY-1); 
		box->box(FL_FRAME_BOX);
		box->align(FL_ALIGN_CENTER|FL_ALIGN_INSIDE|FL_ALIGN_WRAP);
		box->label("Choose LAS-format file\nwith well-log data");
		box->labelsize(18);
		// Initialize the file chooser
		fc = new Fl_Native_File_Chooser();
		fc->filter("ASCII text files \t*.LAS, *.las\n");			 
		end();
	}
};
/////////////////////////////////////////////////////////////////////
//////						CURVE  DRAWING					/////////
//////				in a double-buffered window				/////////
/////////////////////////////////////////////////////////////////////
class Draw_curve : public Fl_Double_Window 
{
	
public:
  	Draw_curve(int X,int Y,int W,int H) : Fl_Double_Window(X,Y,W,H) {}
	
	void draw(void)  // Function draw() within class Draw_seis
	{
		Fl_Double_Window::draw();
		/**    don't declare reserved 'x' and 'y' variables !!!   **/
		int i,j,k;
		int xx, yy, x0, y0, x_dim, y_dim;
		float val,dx,dy,ddy;
		
		//		MARGIN=60;    RIM=10;
		x_dim = w() - (MARGIN+RIM);  y_dim = h() - (MARGIN+RIM);	
		//	x()=y()=0;    w()=268, h()=1018;    Xdim=200, Ydim=950
		//cout << "X = " << x() << ", Y = " << y() << endl;
		//cout << "Width = " << w() << ", Height = " << h() << endl;
		//cout << "Xdim = " << x_dim << ", Y_dim = " << y_dim << endl;		
		/////////////////////////////////////////////////////////////////
		fl_color(51);  fl_rectf(0,0,w(),h());			// Gray area
		fl_color(FL_BLACK); fl_rect(0,0,w(),h());		// bounding rect
		///////    Finally, clean the WORK AREA with WHITE	
		fl_rectf(MARGIN,MARGIN,x_dim,y_dim,255,255,255);
		
		//   G L O B A L variables 
		samp_sta = (int)((dep_sta-depth_sta)/depth_step);
		samp_fin = (int)((dep_end-depth_sta)/depth_step);
		samp_vis = samp_fin-samp_sta;
		
		dx = (float)x_dim/(max_val-min_val);

		dy = dep_end-dep_sta;
		if(dy>5000.) ddy = 1000.0;
		else if(dy>3000.) ddy = 500.0;
		else if(dy>1500.) ddy = 200.0;
		else if(dy>500.) ddy = 100.0;
		else if(dy>100.) ddy = 50.0;
		else if(dy>50.) ddy = 10.0;
		else if(dy>15.) ddy = 5.0;
		else ddy = 1.0;
		
		//////////////////////////////////////////////////////
		//  Draw tick marks and labels on the Depth axis
		fl_line_style(0,1);	// solid, 1 pixel wide
		fl_font(1,11);			// Helvetica, 11 pixel		
		for(val=dep_sta;val<=dep_end;val+=ddy)
		{
			yy = MARGIN + (int)((val-dep_sta)*y_dim/dy);
			fl_color(23); fl_xyline(MARGIN, yy, w()-RIM);	// gray horiz line
			fl_color(0);  fl_xyline(MARGIN-6, yy, MARGIN);	// horiz line
			sprintf(cgbuf,"%g",val);	fl_draw(cgbuf,24,yy+4);
		}
		//////////////////////////////////////////////////////
		//  Draw tick marks and labels on the X-axis
		for(i=0;i<5;i++)
		{
			xx = MARGIN + x_dim*i/4;
			yy = MARGIN-20;
			fl_color(23);	fl_yxline(xx, MARGIN, h()-RIM);		// gray vert. line
			fl_color(0);	fl_yxline(xx, MARGIN-6, MARGIN);	// vertical line
			val = min_val + (max_val-min_val)*i/4.0;
			sprintf(cgbuf,"%g",val);	fl_draw(cgbuf,xx-8,MARGIN-12);
		}
		//////////////////////////////////////////////////////
		//  Draw labels on the coordinate axes		
		sprintf(cgbuf,"Depth,  %s",unit[dep_unit]);
		fl_font(1,12);		fl_draw(90,cgbuf,18,y_dim/2);	// rotated 90˚
		sprintf(cgbuf,"%s",curve[curve_no]+3);
		fl_draw(cgbuf,MARGIN+x_dim/3,25);		
		
		//////////////////////////////////////////////////////
		//  Draw Data Points with optional connection lines
		//////////////////////////////////////////////////////
		if(samp_sta<0) x0 = 0; 
		else {	val = logs[Ncur*samp_sta+curve_no];
					x0 = MARGIN + (int)((val - min_val)*dx); }	
		y0 = MARGIN;
		fl_color(colr);		
		for(i=0;i<samp_vis;i++)
		{
			j = samp_sta+i;
			if(j<0) continue;
			if(j>=Ndat) break;
			k = Ncur*j+curve_no;
			val = logs[k];
			if(val<=no_value) continue;	// ?
				
			xx = MARGIN + (int)((val - min_val)*dx);
			if(xx<MARGIN) xx=MARGIN;
			if(xx>=(MARGIN+x_dim)) xx=MARGIN+x_dim;
			yy = MARGIN + i*y_dim/samp_vis;

			if(pt_size>0)
				fl_pie(xx-pt_size/2,yy-pt_size/2,pt_size,pt_size,0.,360.);
			
			if(line_draw==1) fl_line(x0,y0,xx,yy);
			x0 = xx;		y0 = yy;
			
		}
		
		fl_color(FL_BLACK); fl_rect(MARGIN,MARGIN,x_dim,y_dim);		
		
	} 	// void draw(void)
	
} *curve_plot;
//
/////////////////////////////////////////////////////////////////////
//////					HISTOGRAM  DRAWING					/////////
//////				in a double-buffered window				/////////
/////////////////////////////////////////////////////////////////////
class Draw_hist : public Fl_Double_Window 
{
	
public:
  	Draw_hist(int X,int Y,int W,int H) : Fl_Double_Window(X,Y,W,H) {}
	
	void draw(void)
	{
		Fl_Double_Window::draw();

		int i,j,count,Nsamp;
		int xx, xx1, ww, yy, yy1, hh, x_dim, y_dim;
		int bin_num=HYST_BINS;
		float bin_wid,hist[HYST_BINS];
		float depth,val,Ymin,Ymax,max,scale;
		double pdf, q, sum, valx;
		double coef = sqrt(2.0*M_PI);
		
		x_dim = w() - (MARGIN+RIM);  y_dim = h() - (MARGIN+RIM);	
		/////////////////////////////////////////////////////////////
		fl_color(51);  fl_rectf(0,0,w(),h());			// Gray area
		fl_color(FL_BLACK); fl_rect(0,0,w(),h());		// bounding rect
		///////    Finally, clean the WORK AREA with WHITE	
		fl_rectf(MARGIN,RIM,x_dim,y_dim,255,255,255);
		
		/////////////////////////////////////////////////////////////
		
		bin_wid = (float)((max_val-min_val)/bin_num);
		
		Nsamp=0;
		for(i=0;i<bin_num;i++)				// 1.  Calculate Histogram
		{
			Ymin = min_val + bin_wid*i;
			Ymax = Ymin + bin_wid;
			count = 0;
			for(j=0;j<Ndat;j++)
			{
				depth = logs[Ncur*j];
				if(depth<dep_sta) continue;
				if(depth>=dep_end) continue;				
				val = logs[Ncur*j+curve_no];
				if(val<=no_value) continue;
				if(val<Ymin) continue;
				if(val>=Ymax) continue;
				count++;	Nsamp++;	// Nsamp == total number of measurements
			}							// count == number of measurements in bin 'i'
			hist[i] = (float)count;
		}

		max = 0.0;
		for(i=0;i<bin_num;i++)				// 2.  Normalize Histogram
		{
			val = hist[i]/(float)Nsamp;	// Relative Frequency or Probability
			hist[i] = val;			
			if(val>max) max=val;
		}
		if(max<=0.0)
		{ fl_message("WARNING: No data found!\n\
Close the window to continue...");   return; }
		
		//  Round up Maximum Probability value (scale)
		if(max<.001) scale=.001;		
		else if(max<.005) scale=.005;
			else if(max<.01) scale=.01;
				else if(max<.05) scale=.05;
					else if(max<.1) scale=.1;
						else if(max<.5) scale=.5;
							else scale=1.0;
		
		////////////////////////////////////////////////////////
		yy1 = h() - MARGIN;	
		fl_line_style(0,1);	fl_font(1,12);  fl_color(FL_BLACK);
		///////////////////////////////////////////////////////
		//  X-axis ticks and labels
		for(i=0;i<10;i++)
		{
			val = min_val + (max_val-min_val)*i/10.0;
			xx = MARGIN + i*x_dim/10;
			fl_color(51);	fl_yxline(xx, RIM, yy1);  // light gray
			fl_color(0);	fl_yxline(xx, yy1, yy1+6);
			if(i%2==0)
			{ sprintf(cgbuf,"%g",val); fl_draw(cgbuf,xx-8,yy1+20); }
		}				
	
		for(i=0;i<bin_num;i++)				// 2.  Plot Histogram
		{
			xx = MARGIN + i*x_dim/bin_num;
			xx1 = MARGIN + (i+1)*x_dim/bin_num;
			ww = xx1 - xx;
			hh = hist[i]*y_dim/scale; // Scaled by 'max'
			yy = yy1 - hh;	// 'y' must be smaller than 'h' in 'rectf'
			fl_color(colr);  fl_rectf(xx, yy, ww, hh);
		}

		//////////////////////////////////////////////////////
		//  Y-axis ticks and labels
		fl_color(FL_BLACK);
		for(j=0;j<6;j++)
		{
			val = (100.0 - 20.0*j)*scale;	// Probability in percent
			yy = RIM + j*y_dim/5;
			fl_xyline(MARGIN-6, yy, MARGIN);	// horiz line
			sprintf(cgbuf,"%5.2f",val);	fl_draw(cgbuf,25,yy+4);
		}		
		fl_font(1,14);
		sprintf(cgbuf,"%s",curve[curve_no]+3);
		fl_draw(cgbuf,x_dim*9/10,yy1+36);		
		fl_draw(90,"Probability [ % ]",18,y_dim/3);
		fl_font(0,12);
		sprintf(cgbuf,"Depth Range =  %.0f - %.0f %s",
				  dep_sta,dep_end,unit[dep_unit]);
		fl_draw(cgbuf,MARGIN+x_dim/2,RIM+16);		
		sprintf(cgbuf,"Total number of measurements = %d",Nsamp);
		fl_draw(cgbuf,MARGIN+x_dim/2,RIM+32);
		sprintf(cgbuf,"Number of bins =  %d",bin_num);
		fl_draw(cgbuf,MARGIN+40,RIM+16);		
		sprintf(cgbuf,"Bin width = %g",bin_wid);
		fl_draw(cgbuf,MARGIN+40,RIM+32);		
		
		fl_rect(MARGIN,RIM,x_dim,y_dim);
		
	} 	// void draw(void)
	
} *histo_plot;
//
/////////////////////////////////////////////////////////////////////
//////					CROSS-PLOT							/////////
//////      in a resizable double-buffered window			/////////
/////////////////////////////////////////////////////////////////////
class Draw_cross : public Fl_Double_Window 
{
	
public:
  	Draw_cross(int X,int Y,int W,int H) : Fl_Double_Window(X,Y,W,H) {}
	
	void draw(void)  // Function draw() within class Draw_seis
	{
		Fl_Double_Window::draw();
		/**      don't declare reserved 'x' and 'y' variables !!!    **/		
		int i,j,k;
		int xx, yy, x_dim, y_dim;
		int samp_vis,samp_sta,samp_fin;
		short rgb[3];
		float mk1,mk2, val1,val2,val3,val4, dval1,dval2,dval3,dval4;

		x_dim = w() - (2*MARGIN+RIM);  y_dim = h() - (MARGIN+RIM);			
		//cout << "Xdim = " << x_dim << ", Y_dim = " << y_dim << endl;		
		////////////////////////////////////////////////////////////		
		fl_color(51);  fl_rectf(0,0,w(),h());			// Gray area
		fl_color(FL_BLACK); fl_rect(0,0,w(),h());		// bounding rect
		///////    Finally, clean the WORK AREA with WHITE	
		fl_rectf(MARGIN,RIM,x_dim,y_dim,255,255,255);

		//   G L O B A L variables 
		samp_sta = (int)((dep_sta-depth_sta)/depth_step);
		samp_fin = (int)((dep_end-depth_sta)/depth_step);
		samp_vis = samp_fin-samp_sta;
			
		dval1 = max1 - min1;		dval2 = max2 - min2;
		dval3 = max3 - min3;		dval4 = max4 - min4;
		
		if(dval1<0.01) mk1=0.001;	// labeling interval for cur1
		else if(dval1<0.1) mk1 = 0.01;
		else if(dval1<0.2) mk1 = 0.02;
		else if(dval1<1.0) mk1 = 0.1;
		else if(dval1<2.0) mk1 = 0.2;
		else if(dval1<10.0) mk1 = 1.0;
		else if(dval1<50.0) mk1 = 5.0;
		else if(dval1<100.0) mk1 = 10.0;
		else if(dval1<500.0)	mk1 = 50.0;
		else if(dval1<1000.0) mk1 = 100.0;
		else if(dval1<5000.0) mk1 = 500.0;
		else if(dval1<10000.0) mk1 = 1000.0; else mk1 = 2000.0;
		
		if(dval2<0.01) mk2=0.001;	// labeling interval for cur2
		else if(dval2<0.1) mk2 = 0.01;
		else if(dval2<0.2) mk2 = 0.02;		
		else if(dval2<1.0) mk2 = 0.1;
		else if(dval2<2.0) mk2 = 0.2;						
		else if(dval2<10.0) mk2 = 1.0;
		else if(dval2<50.0) mk2 = 5.0;		
		else if(dval2<100.0) mk2 = 10.0;
		else if(dval2<500.0)	mk2 = 50.0;		
		else if(dval2<1000.0) mk2 = 100.0;
		else if(dval2<5000.0) mk2 = 500.0;		
		else if(dval2<10000.0) mk2 = 1000.0; else mk2 = 2000.0;
		
		////////////////       Cross-Plot      ////////////////////
		//  val1 - X-axis,  val2 - Y-axis,  val3 - color,  val4 - filter		
		// fl_line_style(0,pt_size);	// to use with fl_circle
		for(i=0;i<samp_vis;i++)			// for data samples
		{
			k = samp_sta+i;
			if(k<0) continue;
			if(k>=Ndat) break;			
			val4 = logs[Ncur*k+cur4];
			if(val4>max4 || val4<min4) continue;	// data filtering
			
			val1 = logs[Ncur*k+cur1]; 
			val2 = logs[Ncur*k+cur2];
			if(val1>max1 || val1<min1) continue;
			if(val2>max2 || val2<min2) continue;
			
			xx = MARGIN + (int)((val1-min1)*x_dim/dval1);
			yy = RIM + y_dim - (int)((val2-min2)*y_dim/dval2);
			
			val3 = (logs[Ncur*k+cur3]-min3)/dval3;
			if(val3>1.0) val3 = 1.0;
			if(val3<0.0) val3 = 0.0;
			
			pal_interp(pal,val3,rgb);
			
			// NOTE:   Fl_Color c = fl_rgb_color(rgb[0],rgb[1],rgb[2]);
			fl_color(fl_rgb_color(rgb[0],rgb[1],rgb[2]));
			fl_pie(xx-pt_size/2, yy-pt_size/2, pt_size, pt_size, 0.0, 360.0);
		}
		fl_color(FL_BLACK); fl_rect(MARGIN,RIM,x_dim,y_dim); // bounding rect
		///////////////////////////////////////////////////////////
		
		fl_line_style(0,1);					// solid, 1 pixel wide
		fl_color(FL_BLACK);  fl_font(1,12); // Helvetica, 14 pixel
				
		////     Ticks and labels for the X-axis		////
		fl_draw(curve[cur1]+5,x_dim,h()-25);
		for(val1=min1;val1<=(max1-mk1/10.0);val1+=mk1)
		{
			xx = MARGIN + (val1-min1)*x_dim/dval1;
			yy = RIM + y_dim;
			sprintf(cgbuf,"%g",val1); 
			fl_draw(cgbuf,xx-8,yy+20);		// text string
			fl_yxline(xx, yy, yy+6);		// vertical line			
		}
		
		////     Ticks and labels for the Y-axis		////				
		fl_draw(90,curve[cur2]+5,16,y_dim/4);  // rotated 90˚
		for(val2=min2;val2<(max2-mk2/10.0);val2+=mk2)
		{
			xx = MARGIN;
			yy = RIM + y_dim - (int)((val2-min2)*y_dim/dval2);
			sprintf(cgbuf,"%g",val2);
			fl_draw(cgbuf,32,yy+4);			// text string	
			fl_xyline(xx-6,yy,xx);			// horiz line		
		}
				
		////			Draw Color Map			////
		samp_fin = 125;	// height [pixels] for the color map
		xx = MARGIN+x_dim;
		fl_rect(xx,y_dim/3,18,samp_fin+2); // bounding rect
		
		for(j=0,yy=y_dim/3+1;j<samp_fin;j++,yy++)
		{
			val1 = (float)j/samp_fin;
			pal_interp(pal,val1,rgb);
			fl_rectf(xx+1,yy,16,1,rgb[0], rgb[1], rgb[2]);
		}

		////			Label Color Map			////		
		fl_color(FL_BLACK); // fl_font(1,12); fl_line_style(0,1);
		fl_draw(90,curve[cur3]+5,xx+35,y_dim/3-12);
		
		xx = MARGIN+x_dim+17;
		for(i=0;i<=5;i++)
		{
			yy = y_dim/3 + 1 + i*samp_fin/5;
			val1 = min3 + dval3*i/5.0;	
			fl_xyline(xx,yy,xx+5);
			sprintf(cgbuf,"%g",val1); 
			fl_draw(cgbuf,xx+9,yy+4);
		}				
		
	} 	// void draw(void)	
	
} *cross_plot;
///////////////////////  END of XPLOT  ////////////////////////
//
class Draw_pal : public Fl_Widget  ////  Drawing Color Palette
{
	void draw() 
	{
		int i,j,xx,yy,x_sta,wid,hei;
		short rgb[3];
		float yi;
		
		x_sta = x();  yy = y();  wid = w();  hei = h();
		fl_rectf(x_sta-2,yy-2,wid+4,hei+4,0,0,0); // Black frame
		
		for(i=0,xx=x_sta;i<wid;i++,xx++)
		{	
			yi=pal[0]+i*(pal[PALSZ-1]-pal[0])/wid;	// actually, yi=i/wid
			pal_interp(pal, yi ,rgb);
			fl_rectf(xx,yy,1,hei,rgb[0], rgb[1], rgb[2]);
		}
	} 
	
public:
	Draw_pal(int X,int Y,int W,int H) : Fl_Widget(X,Y,W,H) {}
	
} *palette;
//
////////////////////////////////////////////////////////////////
/************    List of Call Back functions    ***************/
////////////////////////////////////////////////////////////////

void input_curve(Fl_Widget*,void*data) {
	Fl_Input_Choice *in_curve=(Fl_Input_Choice *)data;

	char buf[4];
	strncpy(buf,in_curve->value(),3);
	sscanf(buf,"%d", &curve_no);
	
	//if(curve_no==0) min_val=max_val=0.0;
	//else { min_val=mini[curve_no];  max_val=maxi[curve_no]; }
	min_val=mini[curve_no];			 max_val=maxi[curve_no];
 	sprintf(cgbuf," %g",min_val);  min_inp->value(cgbuf);
	sprintf(cgbuf," %g",max_val);  max_inp->value(cgbuf);	
}
void input_min(Fl_Widget *, void *) {
	min_val = atof(min_inp->value());
}
void input_max(Fl_Widget *, void *) {
	max_val = atof(max_inp->value());
}
void input_dep_sta(Fl_Widget *, void *) {
	dep_sta = atof(dep_sta_inp->value());
	//if(dep_sta<depth_sta) dep_sta=depth_sta;
}
void input_dep_end(Fl_Widget *, void *) {
	dep_end = atof(dep_end_inp->value());
	//if(dep_end>depth_end) dep_end=depth_end;	
}
///////////////////////////////////////////////////
void input_cur1(Fl_Widget*,void*data) {
	Fl_Input_Choice *in_cur1=(Fl_Input_Choice *)data;
	
	char buf[4];
	strncpy(buf,in_cur1->value(),3);
	sscanf(buf,"%d", &cur1);
	
  min1=mini[cur1];	max1=maxi[cur1];
  sprintf(cgbuf," %g",min1);  min1_inp->value(cgbuf);
  sprintf(cgbuf," %g",max1);  max1_inp->value(cgbuf);	
}
//
void input_cur2(Fl_Widget*,void*data) {
	Fl_Input_Choice *in_cur2=(Fl_Input_Choice *)data;
	
	char buf[4];
	strncpy(buf,in_cur2->value(),3);
	sscanf(buf,"%d", &cur2);
	
	min2=mini[cur2];	max2=maxi[cur2];
	sprintf(cgbuf," %g",min2);  min2_inp->value(cgbuf);
	sprintf(cgbuf," %g",max2);  max2_inp->value(cgbuf);	
}
//
void input_cur3(Fl_Widget*,void*data) {
	Fl_Input_Choice *in_cur3=(Fl_Input_Choice *)data;
	
	char buf[4];
	strncpy(buf,in_cur3->value(),3);
	sscanf(buf,"%d", &cur3);
	
	min3=mini[cur3];	max3=maxi[cur3];
	sprintf(cgbuf," %g",min3);  min3_inp->value(cgbuf);
	sprintf(cgbuf," %g",max3);  max3_inp->value(cgbuf);	
}
//
void input_cur4(Fl_Widget*,void*data) {
	Fl_Input_Choice *in_cur4=(Fl_Input_Choice *)data;
	
	char buf[4];
	strncpy(buf,in_cur4->value(),3);
	sscanf(buf,"%d", &cur4);
	
	min4=mini[cur4];	max4=maxi[cur4];
	sprintf(cgbuf," %g",min4);  min4_inp->value(cgbuf);
	sprintf(cgbuf," %g",max4);  max4_inp->value(cgbuf);	
}
//
void input_min1(Fl_Widget *, void *) {
	min1 = atof(min1_inp->value());
}
void input_max1(Fl_Widget *, void *) {
	max1 = atof(max1_inp->value());
}
//
void input_min2(Fl_Widget *, void *) {
	min2 = atof(min2_inp->value());
}
void input_max2(Fl_Widget *, void *) {
	max2 = atof(max2_inp->value());
}
//
void input_min3(Fl_Widget *, void *) {
	min3 = atof(min3_inp->value());
}
void input_max3(Fl_Widget *, void *) {
	max3 = atof(max3_inp->value());
}
//
void input_min4(Fl_Widget *, void *) {
	min4 = atof(min4_inp->value());
}
void input_max4(Fl_Widget *, void *) {
	max4 = atof(max4_inp->value());
}
//
void spinner_cb(Fl_Widget *w, void *) {
   Fl_Spinner *s = (Fl_Spinner *)w;	
   pt_size = (int)s->value();
}
//
void connect_cb(Fl_Widget *w, void *) {
   Fl_Check_Button *check = (Fl_Check_Button *)w;
   line_draw = check->value();
}
///////    Color Chooser    ////////////
void do_color(Fl_Widget*, void*)
{	
	colr = fl_show_colormap(bg_color);
	//canvas->redraw();
}
///////    Palette Chooser    //////////
void choice_pal(Fl_Widget*,void*data) 
{
	Fl_Input_Choice *in_pal=(Fl_Input_Choice *)data;
	
	int num;
	
	strncpy(c_word,in_pal->value(),3);
	sscanf(c_word,"%d", &num);
	switch(num)
	{  
		case 0:	pal = RGB_pal; break;		
		case 1:	pal = BGR_pal; break;
		case 2:	pal = BYD_pal; break;
		case 3:	pal = MGR_pal; break;			
		case 4:	pal = BWR_pal; break;				
		default: break;
	}	
	palette->redraw();	
}
//
///////////    EXIT    ////////////
void exit_b(Fl_Widget *) 
{
	fclose(in_fp);           exit(0);
}
/////////////////////////////////////////////////////////////////////
///////////////           DRAWING Window          ///////////////////
/////////////////////////////////////////////////////////////////////
void doPLOT(Fl_Widget *, void*)
{  

	// Define Window Size with Margins
	win_w = MARGIN+RIM+200;
	win_h = Fl::h() - 40;

	curve_window = new Fl_Double_Window(600,0,win_w,win_h,
													"Curve Display");
	curve_plot = new Draw_curve(0,0,win_w,win_h);
	
	curve_window->end();
	//curve_window->resizable(seis_window);	
	curve_window->show();
	
}
//
void doHIST(Fl_Widget *, void*)
{  
	// Define Window Size with Margins
	hist_w = 572;
	hist_h = 472;
	
	histo_window = new Fl_Double_Window(600,200,hist_w,hist_h,
												"Histogram Window");
	histo_plot = new Draw_hist(0,0,hist_w,hist_h);
	
	histo_window->end();
	//curve_window->resizable(seis_window);	
	histo_window->show();
	
}
/////////////////////////////////////////////////////////////////////
void XPLOT(Fl_Widget *, void*)
{  
	
	// Define Window Size with Margins
	xplot_w = Fl::w()/4;
	xplot_h = Fl::h()*2/5;
		
	X_window = new Fl_Double_Window(600,400,xplot_w,xplot_h,
											"Cross-plot Window");
	cross_plot = new Draw_cross(0,0,xplot_w,xplot_h);
	
	X_window->end();
	X_window->resizable(X_window);	
	X_window->show();
}
//
/////////////////////////////////////////////////////////////////////
/****   MAIN   MAIN   MAIN   MAIN   MAIN   MAIN   MAIN   MAIN   ****/
/////////////////////////////////////////////////////////////////////
int main(int argc, char** argv)
{
	int i,cur;
	double t,val;
	const char *unit[] = {"meter","foot"};
	char cbuf1[300],cbuf2[99],cbuf3[99],cbuf4[99],cbuf5[99],cbuf[80];
	///////////////////////////////////////////////////////////
	
	Fl::scheme(NULL); //  "plastic" for Aqua and "gtk+"
	// NULL use the scheme defined in the FLTK_SCHEME env.var. 
	
	if(argc == 2) strcpy(inpname,argv[1]);		// Do not use terminal
	else		// if you want to create an Apple  *.app bundle application
	{
		File_menu *choice = new File_menu();
		choice->show();
		choice->end(); // no more widgets will be added to the window
		while(chosen<0) Fl::wait();
		choice->hide();
	}
	
	read_file_info(inpname);
	
	opts_window = new Fl_Window(1050,50,420,900,"Display Manager");

	Fl_Box* label1 = new Fl_Box(FL_FRAME_BOX,10,4,400,106,"");
	label1->color(6);
	
	Fl_Box* label2 = new Fl_Box(80,12,250,20,"  File Information  ");
	label2->labelcolor(FL_BLACK);
	label2->labelfont(FL_HELVETICA_BOLD);	label2->labelsize(16);

	sprintf(cbuf1,"Log file contains %d curves,  %d samples each\n\
Starting depth = %.3f     Ending depth = %.3f\n\
Depth increment = %.3f   Unit of measurement = %s\n\
Gaps in observations are filled with   %.3f",
Ncur,Ndat,depth_sta,depth_end,depth_step,unit[dep_unit],no_value);
	Fl_Box* label3 = new Fl_Box(25,22,360,90,cbuf1);
	label3->labelfont(FL_HELVETICA);	label3->labelsize(14);
	

	cur=1;
	Fl_Input_Choice in_curve(225,125,176,24,"Select curve by its mnemonic: ");
	in_curve.callback(input_curve, (void*)&in_curve);
	
	for(i=0;i<Ncur;i++) in_curve.add(curve[i]);
	in_curve.value(cur);  // default curve for plotting
	in_curve.labelsize(14); in_curve.labelfont(FL_HELVETICA_BOLD);
	in_curve.align(Fl_Align(FL_ALIGN_LEFT));	
	

	Fl_Box* label4 = new Fl_Box(40,170,110,24,"Depth range to plot:");
	label4->labelfont(FL_HELVETICA_BOLD); label4->labelsize(14);
	
	dep_sta_inp = new Fl_Float_Input(225, 170, 75, 24, " from ");
	sprintf(cbuf," %g",dep_sta);  
	dep_sta_inp->value(cbuf);   dep_sta_inp->callback(input_dep_sta);
	dep_end_inp = new Fl_Float_Input(325, 170, 75, 24, " to ");
	sprintf(cbuf," %g",dep_end);  
	dep_end_inp->value(cbuf);   dep_end_inp->callback(input_dep_end);
	

	Fl_Box* label5 = new Fl_Box(40,210,120,24,"Range of curve values:");
	label5->labelfont(FL_HELVETICA_BOLD); label5->labelsize(14);
	
	min_inp = new Fl_Float_Input(225, 210, 75, 24, " from ");
	sprintf(cbuf," %g",min_val);  
	min_inp->value(cbuf);   min_inp->callback(input_min);
	max_inp = new Fl_Float_Input(325, 210, 75, 24, " to ");
	sprintf(cbuf," %g",max_val);  
	max_inp->value(cbuf);   max_inp->callback(input_max);
	
	
	//////////////////////////////////////////////////////////////////
	////////////////    Dot attributes      //////////////////////////
	
	Fl_Spinner* spin = new Fl_Spinner(125, 250, 36, 24,
												 "Dot size: ");
	spin->labelsize(14);	spin->labelfont(FL_HELVETICA_BOLD);
	spin->minimum(0);		spin->maximum(7);		spin->step(1);
	spin->value(pt_size);	spin->align(Fl_Align(FL_ALIGN_LEFT));
	spin->callback(spinner_cb);
	Fl_Box* label6 = new Fl_Box(165,250,40,24,"pixel");
	label6->labelfont(FL_HELVETICA); label6->labelsize(14);	
	
	choose_color = new Fl_Button(260,250,120,24,"Curve Color");
	choose_color->labelcolor(FL_BLUE);	choose_color->color(134);
	choose_color->labelfont(3);	choose_color->labelsize(14);
	choose_color->box(FL_RSHADOW_BOX);	choose_color->callback(do_color);
	
	connected = new Fl_Check_Button(60,280,120,24,
										" Connect dots with line?");
	connected->labelfont(FL_HELVETICA_BOLD);	connected->labelsize(14);
	connected->value(line_draw);			connected->callback(connect_cb);

	//////////////////////////////////////////////////////////////////
	apply = new Fl_Button(35,315,155,30,"Display the curve");
	apply->color(62);		apply->labelcolor(FL_BLACK);
	apply->labelfont(3);	apply->labelsize(16);	
	apply->box(FL_RSHADOW_BOX);		apply->callback(doPLOT);
			
	histo = new Fl_Button(225,315,155,30,"Display histogram");
	histo->color(62);						histo->labelcolor(FL_BLACK);
	histo->labelfont(3);					histo->labelsize(16);	
	histo->box(FL_RSHADOW_BOX);		histo->callback(doHIST);

	//////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////
	
	Fl_Box* line1 = new Fl_Box(FL_FRAME_BOX,10,360,400,30,
										"Select the data to cross-plot");
	line1->color(6);					line1->labelcolor(FL_BLACK);
	line1->labelfont(FL_COURIER_BOLD);	line1->labelsize(20);
	
	
	cur1=1;
	Fl_Input_Choice in_cur1(210,400,176,24,"Select curve for X-axis: ");
	in_cur1.callback(input_cur1, (void*)&in_cur1);
	
	for(i=0;i<Ncur;i++) in_cur1.add(curve[i]);
	in_cur1.value(cur1);  // default curve for plotting
	in_cur1.labelsize(15); in_cur1.labelfont(FL_HELVETICA_BOLD);
	in_cur1.align(Fl_Align(FL_ALIGN_LEFT));	
	
	Fl_Box* range1 = new Fl_Box(50,430,120,24,"Range of values to plot:");
	range1->labelfont(FL_HELVETICA_BOLD); range1->labelsize(14);
	
	min1_inp = new Fl_Float_Input(235, 430, 60, 24, " from ");
	sprintf(cbuf," %g",min1);  
	min1_inp->value(cbuf);   min1_inp->callback(input_min1);
	max1_inp = new Fl_Float_Input(325, 430, 60, 24, " to ");
	sprintf(cbuf," %g",max1);  
	max1_inp->value(cbuf);   max1_inp->callback(input_max1);	


	cur2=1;
	Fl_Input_Choice in_cur2(210,480,176,24,"Select curve for Y-axis: ");
	in_cur2.callback(input_cur2, (void*)&in_cur2);
	
	for(i=0;i<Ncur;i++) in_cur2.add(curve[i]);
	in_cur2.value(cur2);  // default curve for plotting
	in_cur2.labelsize(15); in_cur2.labelfont(FL_HELVETICA_BOLD);
	in_cur2.align(Fl_Align(FL_ALIGN_LEFT));	
	
	Fl_Box* range2 = new Fl_Box(50,510,120,24,"Range of values to plot:");
	range2->labelfont(FL_HELVETICA_BOLD); range2->labelsize(14);
	
	min2_inp = new Fl_Float_Input(235, 510, 60, 24, " from ");
	sprintf(cbuf," %g",min2);  
	min2_inp->value(cbuf);   min2_inp->callback(input_min2);
	max2_inp = new Fl_Float_Input(325, 510, 60, 24, " to ");
	sprintf(cbuf," %g",max2);  
	max2_inp->value(cbuf);   max2_inp->callback(input_max2);	


	cur3=1;
	Fl_Input_Choice in_cur3(210,560,176,24,"Curve for color coding: ");
	in_cur3.callback(input_cur3, (void*)&in_cur3);
	
	for(i=0;i<Ncur;i++) in_cur3.add(curve[i]);
	in_cur3.value(cur3);  // default curve for plotting
	in_cur3.labelsize(15); in_cur3.labelfont(FL_HELVETICA_BOLD);
	in_cur3.align(Fl_Align(FL_ALIGN_LEFT));	
	
	Fl_Box* range3 = new Fl_Box(50,590,120,24,"Range of values to use:");
	range3->labelfont(FL_HELVETICA_BOLD); range3->labelsize(14);
	
	min3_inp = new Fl_Float_Input(235, 590, 60, 24, " from ");
	sprintf(cbuf," %g",min3);  
	min3_inp->value(cbuf);   min3_inp->callback(input_min3);
	max3_inp = new Fl_Float_Input(325, 590, 60, 24, " to ");
	sprintf(cbuf," %g",max3);  
	max3_inp->value(cbuf);   max3_inp->callback(input_max3);	
	
	
	cur4=1;
	Fl_Input_Choice in_cur4(210,640,176,24,"Select curve for filtering: ");
	in_cur4.callback(input_cur4, (void*)&in_cur4);
	
	for(i=0;i<Ncur;i++) in_cur4.add(curve[i]);
	in_cur4.value(cur4);  // default curve for plotting
	in_cur4.labelsize(15); in_cur4.labelfont(FL_HELVETICA_BOLD);
	in_cur4.align(Fl_Align(FL_ALIGN_LEFT));	
	
	Fl_Box* range4 = new Fl_Box(50,670,120,24,"Range of values to use:");
	range4->labelfont(FL_HELVETICA_BOLD); range4->labelsize(14);
	
	min4_inp = new Fl_Float_Input(235, 670, 60, 24, " from ");
	sprintf(cbuf," %g",min4);  
	min4_inp->value(cbuf);   min4_inp->callback(input_min4);
	max4_inp = new Fl_Float_Input(325, 670, 60, 24, " to ");
	sprintf(cbuf," %g",max4);  
	max4_inp->value(cbuf);   max4_inp->callback(input_max4);
	
	/////////////////////////////////////////////////////////////////
	
	Fl_Input_Choice in_pal(185,720,180,24,"Choose colormap:");
	in_pal.callback(choice_pal, (void*)&in_pal);
	
	in_pal.add("  0 - Red-Green-Blue   ");
	in_pal.add("  1 - Blue-Green-Red   ");
	in_pal.add("  2 - Blue-Yellow-Dark ");
	in_pal.add("  3 - Magenta-Green-Red");
	in_pal.add("  4 - Blue-Gray-Red    ");
	
	in_pal.value(0);   // default palette number
	in_pal.labelfont(FL_HELVETICA_BOLD);	in_pal.labelsize(15);
	in_pal.align(Fl_Align(FL_ALIGN_LEFT));
	 
	palette = new Draw_pal(90,775,232,16);	// 10*23+2 pixels
	Fl_Box* pal2 = new Fl_Box(80,753,250,20,
					"@<-  LOW          log values          HIGH  @->");		
	pal2->labelfont(FL_HELVETICA_BOLD);	pal2->labelsize(12);
	
	//////////////////////////////////////////////////////////////////

	x_plot = new Fl_Button(100, 825, 100, 30, "Cross-plot");
	x_plot->color(62); x_plot->labelsize(16); x_plot->labelfont(3);
	x_plot->labelcolor(FL_BLACK);		x_plot->box(FL_RSHADOW_BOX);
	x_plot->callback(XPLOT);
	/////////////////////////////////////////////////////////////////
	
	end_this = new Fl_Button(250,825, 80, 30, "Exit");
	end_this->color(208);			end_this->labelcolor(FL_BLACK);
	end_this->labelfont(3);			end_this->labelsize(18);
	end_this->box(FL_RSHADOW_BOX);	end_this->callback(exit_b);

	opts_window->end(); 
	opts_window->show();
	/////////////////////////////////////////////////////////////////
	return(Fl::run());
   return 0;
}
//////////////////////////////////////////////////////////////////////
/****    END_MAIN   END_MAIN   ENF_MAIN   END_MAIN   END_MAIN    ****/
//////////////////////////////////////////////////////////////////////
/****               Called Function Definition                   ****/
//////////////////////////////////////////////////////////////////////
/**
	==============    LAS file format standards    ==================
				Mandatory sections that make up an LAS file:
	~V  - contains version and wrap mode information as following:
 VERS.                 2.00:   CWLS log ASCII Standard -VERSION 2.00
 WRAP.						YES:  Multiple lnes
				or
 WRAP.                  NO:   One line per depth step 
 Non-wrapped line in the data section is no longer than 256 characters  
 
	~W  - contains well identification with mnemonics as indicated:
 STRT.FT                                        0.0000 : Starting Depth
 STOP.FT                                    12820.0000 : Ending Depth
 STEP.FT                                        0.5000 : Level Spacing
 NULL.                                       -999.2500 : Absent value
		( units of depth allowed are FT or M ) 	
	
	~C  - contains curve information
	~P  - contains parameters or constants  (OPTIONAL secion)
	~O  - contains other information			 (OPTIONAL secion)
	~A  - contains ASCII log data with depths in the first column
	
FLAGS are used to assist computers in identifying specific lines in a LAS file:

"~" (tilde) is used to mark the beginning of a section in a file.
It is the first non-space and non-quotation character on a line

"#" is used to indicate a comment line. Can appear only above the data section.

"." and ":" in sections other than the data secton are used to delimit 
information within the line. Info to the right of the colon is a detailed
definition of the mnemonic located to the left of the colon. The dot is used
to separate two mnemonics. Spaces may occur to the right or left of "." or ":".
**/
void read_file_info(const char *name) // Reads ALL log-data into logs[]
{
	int i,j,k,n,pass,counter;
	char *ptr, buf[80], bigbuf[801], keys[]="-1234567890";
	float val;

	if ((in_fp = fopen(name, "r")) == NULL)
	{fl_message("ERROR opening input file!\nClose the window to exit");exit(1);}

	///////////////////////////////////////////////////////////////
	/////    Read LAS sections with mandatory information		//////
	///////////////////////////////////////////////////////////////	
	pass = 0;	fseek(in_fp, 0L, SEEK_SET);	// go to beginning
	while(fgets(bigbuf,800,in_fp))				// while != NULL
	{
		if(strstr(bigbuf,"WRAP") != NULL) // located a string with "WRAP" 
		{
			pass = 1;
			if(strstr(bigbuf,"YES") != NULL) // "YES" in the same string
			{ fl_message("Text wrapping is not supported in this program!\n\
Close the window to exit..."); fclose(in_fp); exit(1); }
			else break;
		}
	}
	if(pass==0)
	{ fl_message("1. LAS mandatory information was not found!\n\
Close the window to exit..."); fclose(in_fp); exit(1); }


	pass = 0;	fseek(in_fp, 0L, SEEK_SET);		// go to beginning
	while(fgets(bigbuf,800,in_fp))				// while != NULL
	{
		if(strstr(bigbuf,"STRT") != NULL) // located a string with "STRT" 
		{
			j=strcspn(bigbuf,".");  dep_unit = 2; // unknown by default
			strncpy(buf,bigbuf+j,2); buf[2]='\0';
			if(buf[1]=='M') dep_unit=0;
			if(buf[1]=='F') dep_unit=1;
		
			i=strcspn(bigbuf,keys); // search for the first digit occurrence
			j=strcspn(bigbuf,":");  // search for the first colon occurrence
			strncpy(buf,bigbuf+i,j-i);
			buf[j]='\0';				// terminating NULL
			sscanf(buf,"%f",&depth_sta);
			
			fgets(bigbuf,800,in_fp);
			if(strstr(bigbuf,"STOP") != NULL) 
			{
				i=strcspn(bigbuf,keys);
				j=strcspn(bigbuf,":");
				strncpy(buf,bigbuf+i,j-i);
				buf[j]='\0';
				sscanf(buf,"%f",&depth_end);			
			}
			else break;
			
			fgets(bigbuf,800,in_fp);					
			if(strstr(bigbuf,"STEP") != NULL) 
			{
				i=strcspn(bigbuf,keys);
				j=strcspn(bigbuf,":");
				strncpy(buf,bigbuf+i,j-i);
				buf[j]='\0';
				sscanf(buf,"%f",&depth_step);			
			}
			else break;
			
			fgets(bigbuf,800,in_fp);					
			if(strstr(bigbuf,"NULL") != NULL) 
			{
				i=strcspn(bigbuf,keys);
				j=strcspn(bigbuf,":");
				strncpy(buf,bigbuf+i,j-i);
				buf[j]='\0';
				sscanf(buf,"%f",&no_value);			
			}
			else break;			
			
			pass=1; break;			
		}
	}
	if(pass==0)
	{ fl_message("2. LAS mandatory information was not found!\n\
Close the window to exit..."); fclose(in_fp); exit(1); }


	///////////////////////////////////////////////////////////////
	/////    Get Info on the Number-of-Curves and Samples 	//////
	///////////////////////////////////////////////////////////////
	pass = 0;	fseek(in_fp, 0L, SEEK_SET);		// go to beginning
	while(fgets(bigbuf,800,in_fp))				// while != NULL
	{
		if(strstr(bigbuf,"~C") != NULL) // located Curve Info Section 
		{
			k=0;
			for(i=0;i<MAX_CURVE;i++)
			{
				fgets(bigbuf,800,in_fp); 
				if(bigbuf[0]=='#') continue;
				if(bigbuf[0]=='~') break;
				
				sprintf(buf,"%3d",k);
				strcat(buf,"  ");  
				strncat(buf,bigbuf,17);
				j=strcspn(buf,"/");    buf[j]='_';	// FLTK doesn,t like "/"
				buf[21]='\0';						
				strncpy(curve[k],buf,22);				
				
				k++; if(k>=MAX_CURVE)
				{ fl_message("ERROR: excessive number of curves\n\
Close the window to exit"); fclose(in_fp); exit(1); }				
			}
			pass=1; Ncur=k; break;		// Number of curves (logs) in file
		}
	}
	if(pass==0)
	{ fl_message("3. LAS mandatory information was not found!\n\
Close the window to exit..."); fclose(in_fp); exit(1); }


	pass=i=0;	fseek(in_fp, 0L, SEEK_SET);		// go to beginning
	while(fgets(bigbuf,800,in_fp))				// while != NULL
	{
		if(pass==1) i++;	
		if(strstr(bigbuf,"~A") != NULL) pass=1; // located Data Section
	}	
	if(pass==0)
	{ fl_message("4. LAS data section  was not found!\n\
Close the window to exit..."); fclose(in_fp); exit(1); }
	
	Ndat = i;		// Number of data samples in each curve (log)
	
	mini = new (nothrow) float[Ncur];			// for logs scaling
	maxi = new (nothrow) float[Ncur];			// for logs scaling
	logs = new (nothrow) float[Ncur * Ndat];	// Allocate MEMORY  ~10 MB
	if(logs == 0)
	{ fl_message("ERROR: NO memory to accomodate log data!\n\
Close the window to exit...");	fclose(in_fp);  exit(1); }


	///////////////////////////////////////////////////////////////
	///////		Position File pointer on a Data Section	  /////////
	fseek(in_fp, 0L, SEEK_SET);
	while(fgets(bigbuf,800,in_fp)) if(strstr(bigbuf,"~A") != NULL) break;
	///////////////////////////////////////////////////////////////
	///////    Read Data Section to the logs[..] array	  /////////
	///////////////////////////////////////////////////////////////			
	i=0;							// i = data sample index
	while(fgets(bigbuf,800,in_fp))
	{
		if(i>=Ndat) 
		{ fl_message("5. Corrupted Log-Data Block\n\
Close the window to exit"); fclose(in_fp); exit(1); }
		ptr = strtok (bigbuf," ");  
		j=0;						// j = curve number index
		while (ptr != NULL)
		{
			sscanf(ptr,"%f",logs+Ncur*i+j);
			ptr = strtok (NULL, " ");	j++;			
		}
		if(j != Ncur) 
		{ fl_message("6. Corrupted Log-Data Block\n\
Close the window to exit"); fclose(in_fp); exit(1); }
		i++;
	}
	//  Final data test...
	//  Most 'Corrupted Log-Data Block' messages are due to strange symbols
	//  (like ?) found at the end-of-file with 'Xcode' editor - delete them!	
	i=(int)logs[0];  k=(int)logs[Ncur*(Ndat-1)];
	j=(int)depth_sta;  n=(int)depth_end;
	if(i == j && k == n) ;
	else { fl_message("7. Corrupted Log-Data Block\n\
Close the window to exit"); fclose(in_fp); exit(1); }

	fclose(in_fp);
	//for(i=0;i<20;i++) printf("i = %d, GR = %.3f\n",i,logs[Ncur*i+8]);	
	
	///////////////////////////////////////////////////////////////
	///////		Estimate the Min-Max of log data		  /////////
	///////////////////////////////////////////////////////////////		
	for(j=0;j<Ncur;j++)				// Search for Min-Max values 
	{
		mini[j]=FLT_MAX;  maxi[j]=FLT_MIN;
		for(i=0;i<Ndat;i++)
		{ 
			val=logs[Ncur*i+j];
			if(val>no_value) 
			{ if(val>maxi[j]) maxi[j]=val; if(val<mini[j]) mini[j]=val; }
		}
		
		val = (maxi[j]-mini[j]);  // Round-up  Min-Max vales
		if(val>1000.0)				
		{ 
			i=(int)(mini[j]/100.0); mini[j]=100.0*i; 
			i=(int)(maxi[j]/100.0); maxi[j]=100.0*(i+1);  continue;
		}
		if(val>100.0) 
		{ 
			i=(int)(mini[j]/10.0); mini[j]=10.0*i; 
			i=(int)(maxi[j]/10.0); maxi[j]=10.0*(i+1);  continue;
		}
		if(val>10.0) 
		{ 
			i=(int)mini[j]; mini[j]=i; 
			i=(int)maxi[j]; maxi[j]=(i+1); continue;
		}
		if(val>1.0) 
		{ 
			i=(int)(mini[j]*10.0); mini[j]=0.1*i; 
			i=(int)(maxi[j]*10.0); maxi[j]=0.1*(i+1); continue;
		}						
		if(val>0.1) 
		{ 
			i=(int)(mini[j]*100.0); mini[j]=0.01*i; 
			i=(int)(maxi[j]*100.0); maxi[j]=0.01*(i+1); continue;
		}
	}

	depth_step = (depth_end-depth_sta)/(Ndat-1);  // More precise..?
	//cout << "Step = " << depth_step << endl;
	dep_sta=depth_sta;  dep_end=depth_end;	// default plotting range
 	min_val=mini[curve_no];		max_val=maxi[curve_no];
 	min1=mini[cur1];		max1=maxi[cur1];	
 	min2=mini[cur2];		max2=maxi[cur2];
 	min3=mini[cur3];		max3=maxi[cur3];
 	min4=mini[cur4];		max4=maxi[cur4];
				
	pal = RGB_pal;							// Default color palette
}
////////////////////////////////////////////////////////////////
//	 Resampling using Cubic Convolution Method [R. Keys, 1981].
// Four data values from an array  dat[0..NS1]  are used every
// time to RETURN a calculated "index" value of the RESAMPLED
// array having size NS2 and same origin as the input array.
////////////////////////////////////////////////////////////////
float conv3(float *dat, int NS1, int NS2, int index)
{
	float y_1,y0,y1,y2,dx,dx2,dx3,k_1,k0,k1,k2;
	
	if(index<0 || index >=NS2) return 0.0;
	
	int i = index*NS1/NS2;
	
	y0=dat[i]; 	
	if(i<1) y_1=y0; else y_1=dat[i-1];
	if(i>NS1-2) y1=y0; else y1=dat[i+1]; 
	if(i>NS1-3) y2=y1; else y2=dat[i+2];
	
	dx = (float)(index*NS1 - i*NS2)/NS2;		
	dx2=dx*dx;  dx3=dx*dx2;		
	
	k_1 = (2.0*dx2 - dx3 - dx)/2.0;
	k0 = (3.0*dx3 - 5.0*dx2 + 2.0)/2.0;
	k1 = (4.0*dx2 - 3.0*dx3 + dx)/2.0;
	k2 = (dx3 - dx2)/2.0;

	return (k_1*y_1 + k0*y0 + k1*y1 + k2*y2);	
}
////////////////////////////////////////////////////////////////
//		Below we assume that all the color palettes have the same
//		range of values from 0.0 to 1.0 with equal PALINC (==0.1)
//
void pal_interp(float palette[], float x, short colors[])
{					//			0.0 <= x <= 1.0 
	int i;
	short r0,g0,b0,dr,dg,db;
	float dx;
	
	r0=g0=b0=dr=dg=db=0;	
	i = (int)(x/PALINC + 0.000001);
	
	if(i<PALSZ && i>=0)		// should be always the case
	{
		dx = (x-palette[i])/PALINC;	
		r0 = palette[PALSZ+i];
		g0 = palette[PALSZ*2+i];
		b0 = palette[PALSZ*3+i];
		
		if(i<(PALSZ-1))
		{
			dr = (short)(dx*(palette[PALSZ+i+1] - r0));
			dg = (short)(dx*(palette[PALSZ*2+i+1] - g0));
			db = (short)(dx*(palette[PALSZ*3+i+1] - b0));		
		}	
	}
	
	colors[0] = r0+dr;
	colors[1] = g0+dg;
	colors[2] = b0+db;
}
